
/*
 * Frank's Akku-Kellermaschine
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "sm.h"
#include "sm.tab.h"

/*
 * Groesse des Speichers
 */

#define SPEICHER	65536

/*
 * Namen der Opcodes
 */

char *opnames[] = {
  "load", "store", "push", "pop", "dup", "swap", "mult", "add", 
  "div",  "sub", "fmult", "fadd", "fdiv", "fsub", "ftoi", "itof",
  "inot", "and", "or", "not", "shl", "shr", "print", "fprint",
  "sprint", "cprint", "read", "fread", "sread", "cread", "jump", "call",
  "bzero", "bnzro", "bpos", "bneg", "ret", "exch", "loadw", "loadb",
  "storew", "storeb"
};

/*
 * Speicher, Stack, Akku und Subroutine Stack
 */

unsigned char *speicher;
int            akku;

#define PC	(0)
#define STACK	(2)
#define SSTCK	(4)

/*
 * Liste mit Labels
 */

lablist *labels = NULL;

/*
 * debug	Debugging beim Ausfuehren
 * loaddebug	Debugging beim Laden
 * asmlog	Wohin mit dem Assemblerlisting
 */

int debug     = 0;
int loaddebug = 0;
FILE *asmlog  = NULL;

/*
 * Von Bison, lex bzw. dem Parser
 */

extern int   yydebug;
extern char  filename[];
extern FILE *yyin;

/*
 * This is non-standard, so I'm defining my own
 */

char *_FP_strdup (char *string)
{
  char *result;

  if (string == NULL)
    return NULL;

  if ((result = (char *) malloc (strlen (string) + 1)) == NULL)
    return NULL;

  strcpy (result, string);
  return result;
}

/*
 * Label definieren
 */

void
AddLabel (char *label, int addr)
{
  lablist *new, *iter=labels;

  while (iter) {
    if (strcmp (iter->label, label) == 0) {
      iter->addr = addr;
      return;
    }
    iter = iter->next;
  }

  if ((new = (lablist *) malloc (sizeof (lablist))) == NULL) {
    fprintf (stderr, "panic: out of memory in AddLabel\n");
    exit    (1);
  }
  new->label = _FP_strdup (label);
  new->addr  = addr;
  new->next  = labels;
  new->refs  = NULL;
  labels     = new;
}

/*
 * Referenziere ein Label
 */

void
RefLabel (char *label, int addr)
{
  lablist *new, *nlab, *iter=labels;

  if ((new = (lablist *) malloc (sizeof (lablist))) == NULL) {
    fprintf (stderr, "panic: out of memory in AddLabel\n");
    exit    (1);
  }
  new->addr  = addr;
  new->refs  = NULL;
  new->label = NULL;
  new->next  = NULL;

  while (iter) {
    if (strcmp (iter->label, label) == 0) {
      new->refs  = iter->refs;
      iter->refs = new;
      return;
    }
    iter = iter->next;
  }

  if ((nlab = (lablist *) malloc (sizeof (lablist))) == NULL) {
    fprintf (stderr, "panic: out of memory in AddLabel\n");
    exit    (1);
  }
  nlab->label = _FP_strdup (label);
  nlab->addr  = 0;
  nlab->refs  = new;
  new->next   = NULL;
  nlab->next  = labels;
  labels      = nlab;
}

/*
 * Aufloesen der Label-Referenzen. Wird nach dem Laden einer Datei
 * aufgerufen. Schaut sich an, welche Labels wo verwendet werden,
 * und ersetzt die Zugriffe durch die tatsaechliche Adresse
 */

void
ResolveAddrs (void)
{
  lablist *iter=labels, *initer;
  unsigned short addr, oaddr;
  int count, flag=0;

  if (loaddebug) 
    fprintf (stderr, "\n");

  while (iter) {
    if (loaddebug) {
      fprintf (stderr, "  %04x: %-20s ... ", iter->addr, iter->label);
    }
    initer = iter->refs;
    count  = 0;

    while (initer) {
      oaddr = GETW (initer->addr);
      addr  = (unsigned short) iter->addr + oaddr;
      PUTW (initer->addr, addr);

      count++;
      initer = initer->refs;
    }
    if (loaddebug) {
      fprintf (stderr, "%3d reference%s.%s", count,
	       (count==1)?"":"s",
	       (count==1)?" ":"");
      if (iter->addr == 0)
	fprintf (stderr, " warning: undefined label\n");
      else
	fprintf (stderr, "\n");
    }
    else if (iter->addr == 0 && count) {
      if (flag++ == 0)
	fprintf (stderr, "\n");
      fprintf (stderr, "warning: %3d reference%s to undefined label %s\n",
	       count, (count==1)?"":"s", iter->label);
    }

    iter = iter->next;
  }
}

/*
 * Loescht die Labellisten aus dem Speicher
 */

void
KillLabels (void)
{
  lablist *iter = labels, *initer, *next;

  while (iter) {
    if ((initer = iter->refs)) {
      while (initer) {
	next = initer->next;
	free (initer);
	initer = next;
      }
    }
    next = iter->next;
    free (iter->label);
    free (iter);
    iter = next;
  }
  labels = NULL;
}

/*
 * Der Simulator
 */

void
Simuliere (void)
{
  static char input[256]={0}, *ip=input, *ep;
  int pc, temp;
  /*
   * Wir sind genau dann fertig, wenn der PC 0 ist.
   */
  while (pc = GETW(PC)) {
    switch (speicher[pc] & 0x03) {
    case 0:
      if (debug) {
	fprintf (stderr, " <%08x:%04x> %04x   %02x %15s %s\n",
		 akku, (int)GETW(STACK), pc, (int) speicher[pc], "",
		 opnames[speicher[pc]>>2]);
      }
      PUTW (PC, pc+1);
      break;
    case 1:
      if (debug) {
	fprintf (stderr, 
		 " <%08x:%04x> %04x   %02x %02x %02x %9s %s $%04x\n",
		 akku, (int)GETW(STACK), pc,
		 (int) speicher[pc],
		 (int) speicher[pc+1],
		 (int) speicher[pc+2],
		 "",
		 opnames[speicher[pc]>>2],
		 (int) GETW (pc+1));
      }
      PUTW (PC, pc+3);
      break;
    case 2:
      if (debug) {
	fprintf (stderr, 
		 " <%08x:%04x> %04x   %02x %02x %02x %02x %02x %3s %s #$%08x\n",
		 akku, (int)GETW(STACK), pc,
		 (int) speicher[pc],
		 (int) speicher[pc+1],
		 (int) speicher[pc+2],
		 (int) speicher[pc+3],
		 (int) speicher[pc+4],
		 "",
		 opnames[speicher[pc]>>2],
		 GETD (pc+1));
      }
      PUTW (PC, pc+5);
      break;
    }

    switch (speicher[pc]) {
    case OPC_LOAD_STK:
      akku   = GETD (GETD(GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_LOAD_MEM:
      akku   = GETD (GETW(pc+1));
      break;
      
    case OPC_LOAD_IMM:
      akku   = GETD (pc+1);
      break;

    case OPC_LOADW_STK:
      akku   = (int) GETW (GETD(GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_LOADW_MEM:
      akku   = (int) GETW (GETW(pc+1));
      break;

    case OPC_LOADB_STK:
      akku   = (int) GETB (GETD(GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_LOADB_MEM:
      akku   = (int) GETB (GETD(pc+1));
      break;

    case OPC_STORE_STK:
      PUTD (GETD(GETW(STACK)-4), akku);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_STORE_MEM:
      PUTD (GETW(pc+1), akku);
      break;

    case OPC_STOREW_STK:
      PUTW (GETD(GETW(STACK)-4), akku);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_STOREW_MEM:
      PUTW (GETW(pc+1), akku);
      break;

    case OPC_STOREB_STK:
      PUTB (GETD(GETW(STACK)-4), akku);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_STOREB_MEM:
      PUTB (GETW(pc+1), akku);
      break;

    case OPC_PUSH_STK:
      PUTD (GETW(STACK), akku);
      PUTW (STACK, GETW (STACK)+4);
      break;

    case OPC_POP_STK:
      akku   = GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_DUP_STK:
      PUTD (GETW(STACK), GETD (GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)+4);
      break;

    case OPC_SWAP_STK:
      temp = GETD (GETW(STACK)-8);
      PUTD (GETW(STACK)-8, GETD(GETW(STACK)-4));
      PUTD (GETW(STACK)-4, temp);
      break;

    case OPC_EXCH_STK:
      temp = akku;
      akku = GETD (GETW(STACK)-4);
      PUTD (GETW(STACK)-4, temp);
      break;

    case OPC_MULT_STK:
      akku  *= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_MULT_MEM:
      akku  *= GETD (GETW (pc+1));
      break;

    case OPC_MULT_IMM:
      akku  *= GETD (pc+1);
      break;

    case OPC_ADD_STK:
      akku  += GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_ADD_MEM:
      akku  += GETD (GETW (pc+1));
      break;

    case OPC_ADD_IMM:
      akku  += GETD (pc+1);
      break;

    case OPC_DIV_STK:
      akku  /= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_DIV_MEM:
      akku  /= GETD (GETW (pc+1));
      break;

    case OPC_DIV_IMM:
      akku  /= GETD (pc+1);
      break;

    case OPC_SUB_STK:
      akku  -= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_SUB_MEM:
      akku  -= GETD (GETW (pc+1));
      break;

    case OPC_SUB_IMM:
      akku  -= GETD (pc+1);
      break;

    case OPC_FMULT_STK:
      *((double *) (speicher+(unsigned short)akku)) =
	*((double *) (speicher+(unsigned short)akku)) *
	  *((double *) (speicher+GETD(GETW(STACK)-4)));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_FADD_STK:
      *((double *) (speicher+(unsigned short)akku)) =
	*((double *) (speicher+(unsigned short)akku)) +
	  *((double *) (speicher+GETD(GETW(STACK)-4)));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_FDIV_STK:
      *((double *) (speicher+(unsigned short)akku)) =
	*((double *) (speicher+(unsigned short)akku)) /
	  *((double *) (speicher+GETD(GETW(STACK)-4)));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_FSUB_STK:
      *((double *) (speicher+(unsigned short)akku)) =
	*((double *) (speicher+(unsigned short)akku)) -
	  *((double *) (speicher+GETD(GETW(STACK)-4)));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_FTOI_STK:
      akku   = (int) *((double *) (speicher+GETD(GETW(STACK)-4)));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_FTOI_MEM:
      akku   = (int) *((double *) (speicher+GETW(pc+1)));
      break;
      
    case OPC_ITOF_STK:
      *((double *) (speicher+GETD(GETW(STACK)-4))) = (double) akku;
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_ITOF_MEM:
      *((double *) (speicher+GETW(pc+1))) = (double) akku;
      break;

    case OPC_INOT_STK:
      akku = ! akku;
      break;

    case OPC_NOT_STK:
      akku   = ~akku;
      break;

    case OPC_AND_STK:
      akku  &= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_AND_MEM:
      akku  &= GETD (GETW (pc+1));
      break;

    case OPC_AND_IMM:
      akku  &= GETD (pc+1);
      break;

    case OPC_OR_STK:
      akku  |= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_OR_MEM:
      akku  |= GETD (GETW (pc+1));
      break;

    case OPC_OR_IMM:
      akku  |= GETD (pc+1);
      break;

    case OPC_SHL_STK:
      akku <<= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_SHL_MEM:
      akku <<= GETD (GETW (pc+1));
      break;

    case OPC_SHL_IMM:
      akku <<= GETD (pc+1);
      break;

    case OPC_SHR_STK:
      akku >>= GETD (GETW(STACK)-4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_SHR_MEM:
      akku >>= GETD (GETW (pc+1));
      break;

    case OPC_SHR_IMM:
      akku >>= GETD (pc+1);
      break;

    case OPC_PRINT_STK:
      fprintf (stdout, "%d", akku);
      fflush  (stdout);
      break;

    case OPC_CPRINT_STK:
      fprintf (stdout, "%c", akku);
      fflush  (stdout);
      break;

    case OPC_FPRINT_STK:
      fprintf (stdout, "%g",
	       *((double *) (speicher + (unsigned short) akku)));
      fflush  (stdout);
      break;

    case OPC_SPRINT_STK:
      fprintf (stdout, "%s",
	       (char *) (speicher + (unsigned short) akku));
      fflush  (stdout);
      break;

    case OPC_READ_STK:
      akku = strtol (ip, &ep, 0);
      while (ip == ep) {
	if (fgets (ip=input, 256, stdin) == 0) {
	  input[0] = '\0';
	  akku = -1;
	  break;
	}
	akku = strtol (ip, &ep, 0);
      }
      if (*ip != '\0' && *ep == '\0')
	akku = 0;
      else {
	ip = ep;
	while (*ip == ' ' || *ip == '\t' || *ip == '\n')
	  ip++;
      }
      break;

    case OPC_CREAD_STK:
      if (*ip == '\0') {
	if (fgets (ip=input, 256, stdin) == 0) {
	  input[0]='\0';
	  akku = -1;
	  break;
	}
      }
      akku = (int) (*ip++);
      break;

    case OPC_FREAD_STK:
      *((double*)(speicher+(unsigned short)akku)) = strtod (ip, &ep);
      while (ip == ep) {
	if (fgets (ip=input, 256, stdin) == 0) {
	  input[0] = '\0';
	  *((double*)(speicher+(unsigned short)akku)) = -1.0;
	  break;
	}
	*((double*)(speicher+(unsigned short)akku)) = strtod (ip, &ep);
      }
      if (*ip != '\0' && *ep == '\0')
	*((double*)(speicher+(unsigned short)akku)) = 0.0;
      else {
	ip = ep;
	while (*ip == ' ' || *ip == '\t' || *ip == '\n')
	  ip++;
      }
      break;

    case OPC_SREAD_STK:
      if (*ip == '\0') {
	if (fgets (ip=input, 256, stdin) == 0) {
	  input[0]='\0';
	  break;
	}
      }
      if (input[strlen(input)-1]=='\n')
	input[strlen(input)-1]='\0';

      strcpy ((char*)(speicher+(unsigned short)akku), ip);
      *(ip=input) = '\0';
      break;

    case OPC_JUMP_STK:
      PUTW (PC, GETD (GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_JUMP_MEM:
      PUTW (PC, GETW (pc+1));
      break;

    case OPC_BZERO_STK:
      if (akku == 0)
	PUTW (PC, GETD (GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_BZERO_MEM:
      if (akku == 0)
	PUTW (PC, GETW (pc+1));
      break;

    case OPC_BNZRO_STK:
      if (akku != 0)
	PUTW (PC, GETD (GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_BNZRO_MEM:
      if (akku != 0)
	PUTW (PC, GETW (pc+1));
      break;

    case OPC_BPOS_STK:
      if (akku > 0)
	PUTW (PC, GETD (GETW(STACK)-4));
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_BPOS_MEM:
      if (akku > 0)
	PUTW (PC, GETW (pc+1));
      break;

    case OPC_BNEG_STK:
	PUTW (PC, GETD (GETW(STACK)-4));
      if (akku < 0)
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_BNEG_MEM:
      if (akku < 0)
	PUTW (PC, GETW (pc+1));
      break;

    case OPC_CALL_STK:
      PUTD (GETW(SSTCK), GETW (PC));
      PUTW (PC,    GETD (GETW(STACK)-4));
      PUTW (SSTCK, GETW (SSTCK)+4);
      PUTW (STACK, GETW (STACK)-4);
      break;

    case OPC_CALL_MEM:
      PUTD (GETW(SSTCK), GETW (PC));
      PUTW (PC,    GETW (pc+1));
      PUTW (SSTCK, GETW (SSTCK)+4);
      break;

    case OPC_RET_STK:
      PUTW (PC,    GETD (GETW(SSTCK)-4));
      PUTW (SSTCK, GETW (SSTCK)-4);
      break;

    default:
      fprintf (stderr, "Illegal Opcode %02x at %04x\n", (int)speicher[pc],pc);
      break;
    }
  }
}

void
main (int argc, char *argv[])
{
  FILE *file=NULL;
  int index;

  strcpy (filename, "<stdin>");

  for (index=1; index<argc; index++) {
    if (*argv[index] == '-') {
      switch (argv[index][1]) {
      case 'l':
	loaddebug = 1;
	asmlog    = stderr;
	break;
      case 'r':
	debug = 1;
	break;
      case 'd':
	yydebug = 1;
	break;
      }
    }
    else if (file==NULL) {
      strcpy (filename, argv[index]);
      if ((file = fopen (argv[index], "r")) == NULL) {
	fprintf (stderr, "error: cannot read file %s\n", argv[index]);
	exit    (1);
      }
    }
  }

  fprintf (stderr, "Initializing ... ");
  fflush  (stderr);

  if ((speicher = (unsigned char *) malloc (SPEICHER)) == NULL) {
    fprintf (stderr, "\npanic: cannot alloc machine's memory\n");
    exit    (1);
  }

  memset (speicher, '\0', SPEICHER  * sizeof (unsigned char));

  fprintf (stderr, "done.\nLoading file ... ");
  fflush  (stderr);

  if (file) {
    yyin = file;
  }

  if (yyparse()) {
    fprintf (stderr, "error: input file erroneous\n");
  }
  else {
    fprintf (stderr, "done.\nResolving labels ... ");
    fflush  (stderr);
    ResolveAddrs ();
    fprintf (stderr, "done.\nRunning ...\n");
    Simuliere ();
  }

  KillLabels   ();
  free (speicher);

  if (file)
    fclose (file);
}
