
/*
 * Typdefinitionen der Maschine
 */

typedef struct _lablist {
  char *label;
  int   addr;
  struct _lablist *refs;
  struct _lablist *next;
} lablist;

/*
 * Speicher und andere Deklarationen
 */

extern unsigned char *speicher;
extern int            loaddebug;
extern char          *opnames[];
extern FILE          *asmlog;

/*
 * Definieren und Referenzieren eines Labels
 */

extern void DefLabel (char *, int);
extern void RefLabel (char *, int);

/*
 * Brauchen wir auf manchen Kisten
 */

extern char *_FP_strdup (char *);

/*
 * Makros fuer Speicherzugriff
 * Da viele Prozessoren es nicht gestatten, words und dwords von nicht
 * auf word bzw. dword grenze ausgerichteten Adressen zu lesen. In dem
 * Falle splitten wir dann den Zugriff auf
 */

/*
 * Grr. Diese Makros sind zwar schoen haesslich, aber sie funktionieren
 * nicht. Durch die Bitfummeleien kann es sein, dass INTS auf Bytegrenze
 * anders dargestellt werden als INTS auf Wort oder DWortgrenze.
 * Das stoert zwar nicht, solange _immer_ mit diesen Makros zugegriffen
 * wird, schafft aber Probleme, wenn z.B. ein Speicherbereich kopiert
 * wird.
 */

/*
#define GETB(a)   (*((unsigned char *) (a)))
#define GETW(a)   (((int)(a)&0x01)?((unsigned short)GETB(a)|((unsigned short)GETB((a)+1)<<8)):(*((unsigned short *)(a))))
#define GETD(a)   (((int)(a)&0x01)?((int)GETB(a)|((int)GETW((a)+1)<<8)|((int)GETB((a)+3)<<24)):(((int)(a)&0x02)?((int)GETW(a)|((int)GETW((a)+2)<<16)):(*((int*)(a)))))

#define PUTB(a,b) (*((unsigned char *)(a))=(unsigned char)(b))
#define PUTW(a,b) (((int)(a)&0x01)?((unsigned short)PUTB(a,(b)&0xff)|((unsigned short)PUTB((a)+1,(b)>>8)<<8)):(*((unsigned short *)(a))=(unsigned short)(b)))
#define PUTD(a,b) (((int)(a)&0x01)?((int)PUTB(a,(b)&0xff)|((int)PUTW((a)+1,((b)>>8)&0xffff)<<8)|((int)PUTB((a)+3,(b)>>24)<<24)):(((int)(a)&0x02)?((int)PUTW(a,(b)&0xffff)|((int)PUTW((a)+2,(b)>>16)<<16)):(*((int*)(a))=(b))))
*/

/*
 * Da machen wir doch lieber alles selber und bleiben bei Big-Endian
 */

#define GETB(a)	  (*(speicher+(unsigned short)(a)))
#define GETW(a)	  ((unsigned short)GETB(a)|((unsigned short)GETB((a)+1)<<8))
#define GETD(a)   ((int)GETW(a)|((int)GETW((a)+2)<<16))

#define PUTB(a,b) (*(speicher+(unsigned short)(a))=(unsigned char)(b))
#define PUTW(a,b) ((unsigned short)PUTB(a,(b)&0xff)|((unsigned short)PUTB((a)+1,(b)>>8)<<8))
#define PUTD(a,b) ((int)PUTW(a,(b)&0xffff)|((int)PUTW((a)+2,(b)>>16)<<16))

/*
 * Die ersten 6 Bits eines Opcodes sind die Befehlsnummer, die letzten zwei
 * die Adressierungsart
 */

#define MKCODE(c,a)	(((c)<<2)|a)

/*
 * Befehlsnummern
 */

#define OPC_LOAD	(0)
#define OPC_STORE	(1)
#define OPC_PUSH	(2)
#define OPC_POP		(3)
#define OPC_DUP		(4)
#define OPC_SWAP	(5)
#define OPC_MULT	(6)
#define OPC_ADD		(7)
#define OPC_DIV		(8)
#define OPC_SUB		(9)
#define OPC_FMULT	(10)
#define OPC_FADD	(11)
#define OPC_FDIV	(12)
#define OPC_FSUB	(13)
#define OPC_FTOI	(14)
#define OPC_ITOF	(15)
#define OPC_INOT	(16)
#define OPC_AND		(17)
#define OPC_OR		(18)
#define OPC_NOT		(19)
#define OPC_SHL		(20)
#define OPC_SHR		(21)
#define OPC_PRINT	(22)
#define OPC_FPRINT	(23)
#define OPC_SPRINT	(24)
#define OPC_CPRINT	(25)
#define OPC_READ	(26)
#define OPC_FREAD	(27)
#define OPC_SREAD	(28)
#define OPC_CREAD	(29)
#define OPC_JUMP	(30)
#define OPC_CALL	(31)
#define OPC_BZERO	(32)
#define OPC_BNZRO	(33)
#define OPC_BPOS	(34)
#define OPC_BNEG	(35)
#define OPC_RET		(36)
#define OPC_EXCH	(37)
#define OPC_LOADW	(38)
#define OPC_LOADB	(39)
#define OPC_STOREW	(40)
#define OPC_STOREB	(41)

/*
 * Adressierungsarten
 */

#define ADR_STK		(0)
#define ADR_MEM		(1)
#define ADR_IMM		(2)

/*
 * Opcodes
 */

#define OPC_LOAD_STK	MKCODE(OPC_LOAD,ADR_STK)
#define OPC_LOAD_MEM	MKCODE(OPC_LOAD,ADR_MEM)
#define OPC_LOAD_IMM	MKCODE(OPC_LOAD,ADR_IMM)

#define OPC_STORE_STK	MKCODE(OPC_STORE,ADR_STK)
#define OPC_STORE_MEM	MKCODE(OPC_STORE,ADR_MEM)

#define OPC_PUSH_STK	MKCODE(OPC_PUSH,ADR_STK)
#define OPC_POP_STK	MKCODE(OPC_POP,ADR_STK)
#define OPC_DUP_STK	MKCODE(OPC_DUP,ADR_STK)
#define OPC_SWAP_STK	MKCODE(OPC_SWAP,ADR_STK)

#define OPC_MULT_STK	MKCODE(OPC_MULT,ADR_STK)
#define OPC_MULT_MEM	MKCODE(OPC_MULT,ADR_MEM)
#define OPC_MULT_IMM	MKCODE(OPC_MULT,ADR_IMM)
#define OPC_ADD_STK	MKCODE(OPC_ADD,ADR_STK)
#define OPC_ADD_MEM	MKCODE(OPC_ADD,ADR_MEM)
#define OPC_ADD_IMM	MKCODE(OPC_ADD,ADR_IMM)
#define OPC_DIV_STK	MKCODE(OPC_DIV,ADR_STK)
#define OPC_DIV_MEM	MKCODE(OPC_DIV,ADR_MEM)
#define OPC_DIV_IMM	MKCODE(OPC_DIV,ADR_IMM)
#define OPC_SUB_STK	MKCODE(OPC_SUB,ADR_STK)
#define OPC_SUB_MEM	MKCODE(OPC_SUB,ADR_MEM)
#define OPC_SUB_IMM	MKCODE(OPC_SUB,ADR_IMM)

#define OPC_FMULT_STK	MKCODE(OPC_FMULT,ADR_STK)
#define OPC_FADD_STK	MKCODE(OPC_FADD,ADR_STK)
#define OPC_FDIV_STK	MKCODE(OPC_FDIV,ADR_STK)
#define OPC_FSUB_STK	MKCODE(OPC_FSUB,ADR_STK)

#define OPC_FTOI_STK	MKCODE(OPC_FTOI,ADR_STK)
#define OPC_FTOI_MEM	MKCODE(OPC_FTOI,ADR_MEM)
#define OPC_ITOF_STK	MKCODE(OPC_ITOF,ADR_STK)
#define OPC_ITOF_MEM	MKCODE(OPC_ITOF,ADR_MEM)

#define OPC_INOT_STK	MKCODE(OPC_INOT,ADR_STK)
#define OPC_AND_STK	MKCODE(OPC_AND,ADR_STK)
#define OPC_AND_MEM	MKCODE(OPC_AND,ADR_MEM)
#define OPC_AND_IMM	MKCODE(OPC_AND,ADR_IMM)
#define OPC_OR_STK	MKCODE(OPC_OR,ADR_STK)
#define OPC_OR_MEM	MKCODE(OPC_OR,ADR_MEM)
#define OPC_OR_IMM	MKCODE(OPC_OR,ADR_IMM)
#define OPC_NOT_STK	MKCODE(OPC_NOT,ADR_STK)
#define OPC_SHL_STK	MKCODE(OPC_SHL,ADR_STK)
#define OPC_SHL_MEM	MKCODE(OPC_SHL,ADR_MEM)
#define OPC_SHL_IMM	MKCODE(OPC_SHL,ADR_IMM)
#define OPC_SHR_STK	MKCODE(OPC_SHR,ADR_STK)
#define OPC_SHR_MEM	MKCODE(OPC_SHR,ADR_MEM)
#define OPC_SHR_IMM	MKCODE(OPC_SHR,ADR_IMM)

#define OPC_PRINT_STK	MKCODE(OPC_PRINT,ADR_STK)
#define OPC_FPRINT_STK	MKCODE(OPC_FPRINT,ADR_STK)
#define OPC_SPRINT_STK	MKCODE(OPC_SPRINT,ADR_STK)
#define OPC_CPRINT_STK	MKCODE(OPC_CPRINT,ADR_STK)
#define OPC_READ_STK	MKCODE(OPC_READ,ADR_STK)
#define OPC_FREAD_STK	MKCODE(OPC_FREAD,ADR_STK)
#define OPC_SREAD_STK	MKCODE(OPC_SREAD,ADR_STK)
#define OPC_CREAD_STK	MKCODE(OPC_CREAD,ADR_STK)

#define OPC_JUMP_STK	MKCODE(OPC_JUMP,ADR_STK)
#define OPC_JUMP_MEM	MKCODE(OPC_JUMP,ADR_MEM)

#define OPC_CALL_STK	MKCODE(OPC_CALL,ADR_STK)
#define OPC_CALL_MEM	MKCODE(OPC_CALL,ADR_MEM)

#define OPC_BZERO_STK	MKCODE(OPC_BZERO,ADR_STK)
#define OPC_BZERO_MEM	MKCODE(OPC_BZERO,ADR_MEM)

#define OPC_BNZRO_STK	MKCODE(OPC_BNZRO,ADR_STK)
#define OPC_BNZRO_MEM	MKCODE(OPC_BNZRO,ADR_MEM)

#define OPC_BPOS_STK	MKCODE(OPC_BPOS,ADR_STK)
#define OPC_BPOS_MEM	MKCODE(OPC_BPOS,ADR_MEM)

#define OPC_BNEG_STK	MKCODE(OPC_BNEG,ADR_STK)
#define OPC_BNEG_MEM	MKCODE(OPC_BNEG,ADR_MEM)

#define OPC_RET_STK	MKCODE(OPC_RET,ADR_STK)

#define OPC_EXCH_STK	MKCODE(OPC_EXCH,ADR_STK)

#define OPC_LOADW_STK	MKCODE(OPC_LOADW,ADR_STK)
#define OPC_LOADW_MEM	MKCODE(OPC_LOADW,ADR_MEM)
#define OPC_LOADB_STK	MKCODE(OPC_LOADB,ADR_STK)
#define OPC_LOADB_MEM	MKCODE(OPC_LOADB,ADR_MEM)

#define OPC_STOREW_STK	MKCODE(OPC_STOREW,ADR_STK)
#define OPC_STOREW_MEM	MKCODE(OPC_STOREW,ADR_MEM)
#define OPC_STOREB_STK	MKCODE(OPC_STOREB,ADR_STK)
#define OPC_STOREB_MEM	MKCODE(OPC_STOREB,ADR_MEM)

