%{
#include <stdlib.h>
#include <string.h>
#include "sm.h"
#include "sm.tab.h"

extern char filename[];
extern int  line;

/*
 * Include stack. Only works with FLEX
 * Lex will give a warning on the <<EOF>> line. Ignore it.
 */

#ifdef FLEX_SCANNER

#define MAX_INCLUDE_DEPTH 10

YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
char            include_name[MAX_INCLUDE_DEPTH][256];
char            include_line[MAX_INCLUDE_DEPTH];
int include_stack_ptr = 0;

#endif

%}

%%
#include\ [A-Za-z0-9_\.\/]+ {
#ifdef FLEX_SCANNER
  if (include_stack_ptr >= MAX_INCLUDE_DEPTH) {
    fprintf (stderr, "error: includes nested too deeply" );
    exit(1);
  }
  include_stack[include_stack_ptr] = YY_CURRENT_BUFFER;
  include_line [include_stack_ptr] = line;
  strcpy (include_name[include_stack_ptr], filename);

  if ((yyin = fopen (yytext+9, "r" )) == NULL) {
    fprintf (stderr, "error: cannot open %s\n", yytext+9);
  }
  else {
    include_stack_ptr++;
    strcpy (filename, yytext+9);
    line = 1;
    yy_switch_to_buffer (yy_create_buffer (yyin, YY_BUF_SIZE));
  }
#else
  fprintf (stderr, "error: includes do not work if compiled with lex.\n");
#endif
}
<<EOF>> {
#ifdef FLEX_SCANNER
  if (--include_stack_ptr < 0) {
    yyterminate();
  }
  else {
    yy_delete_buffer( YY_CURRENT_BUFFER );
    yy_switch_to_buffer(include_stack[include_stack_ptr]);
    line = include_line[include_stack_ptr];
    strcpy (filename, include_name[include_stack_ptr]);
  }
#endif
}
[0-9a-fA-F]+[ \t]*h(ex)? {
  yylval.integer = strtol (yytext, NULL, 16);
  return INTEGER;
}
$[ \t]*[0-9a-fA-F]+ {
  yylval.integer = strtol (yytext+1, NULL, 16);
  return INTEGER;
}
[+-]?[0-9]+ {
  yylval.integer = strtol (yytext, NULL, 10);
  return INTEGER;
}
[+-]?[0-9]*\.[0-9]+ {
  yylval.numeric = strtod (yytext, NULL);
  return NUMERIC;
}
[+-]?[0-9]*\.[0-9]+[eE][+-]?[0-9]+ {
  yylval.numeric = strtod (yytext, NULL);
  return NUMERIC;
}
\"[^"]*\" {
  int si=1, di=0;
  while (yytext[si] != '"') {
    if (yytext[si] == '\\') {
      switch (yytext[si+1]) {
      case '\\': yylval.string[di++] = '\\'; break;
      case 'n':  yylval.string[di++] = '\n'; break;
      case 'r':  yylval.string[di++] = '\r'; break;
      }
      si+=2;
    }
    else
      yylval.string[di++] = yytext[si++];
  }
  yylval.string[di] = '\0';
  return STRING;
}
[A-Z_][0-9a-zA-Z_]* {
  strcpy (yylval.string, yytext);
  return LABEL;
}
[A-Z_][0-9a-zA-Z_]*\: {
  strcpy (yylval.string, yytext);
  yylval.string[strlen(yytext)-1]='\0';
  return LABDEF;
}
align			return ALIGN;
load			return LOAD;
loadw			return LOADW;
loadb			return LOADB;
store			return STORE;
storew			return STOREW;
storeb			return STOREB;
push			return PUSH;
pop			return POP;
dup			return DUP;
exch			return EXCH;
swap			return SWAP;
mult			return MULT;
add			return ADD;
div			return DIV;
sub			return SUB;
fmult			return FMULT;
fadd			return FADD;
fdiv			return FDIV;
fsub			return FSUB;
ftoi			return FTOI;
itof			return ITOF;
inot			return INOT;
and			return AND;
or			return OR;
not			return NOT;
shl			return SHL;
shr			return SHR;
print			return PRINT;
fprint			return FPRINT;
sprint			return SPRINT;
cprint			return CPRINT;
read			return READ;
fread			return FREAD;
sread			return SREAD;
cread			return CREAD;
jump			return JUMP;
call			return CALL;
bzero			return BZERO;
bnzro			return BNZRO;
bpos			return BPOS;
bneg			return BNEG;
ret			return RET;
data			return DATA;
\;[^\n]*		/* comment     */ ;
[ \t]+			/* do nothing  */ ;
\n			/* end-of-line */ line++;
.			return yytext[0];
