%{
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "sm.h"

#define YYDEBUG	1

/*
 * Adresse, an der wir gerade sind
 */

static unsigned short loadaddr;

/*
 * Zeilenzaehler
 */

char filename[256];
int  line;

%}

%union {
  unsigned char opcode;
  struct _t_com {
    unsigned char  opcode;
    int            parameter;
  } com;
  unsigned char  byte;
  unsigned short word;
  int    integer;
  double numeric;
  char   string[256];
  int    none;
}

%token <integer>	INTEGER
%token <numeric>	NUMERIC
%token <string>		STRING LABEL LABDEF
%token <byte>		BYTE
%token <word>		WORD
%token			'x' 'w'

%left			'-' '+'
%left			'*' '/'

%token <opcode>		LOAD STORE PUSH POP DUP SWAP MULT ADD DIV SUB
%token <opcode>		FMULT FADD FDIV FSUB FTOI ITOF INOT AND OR NOT
%token <opcode>		SHL SHR PRINT FPRINT SPRINT CPRINT READ FREAD
%token <opcode>		SREAD CREAD JUMP CALL BZERO BNZRO BPOS BNEG RET
%token <opcode>		EXCH DATA LOADW LOADB STOREW STOREB ALIGN

%type <none>		item items file
%type <com>		befehl
%type <com>		bef_load bef_store bef_push bef_pop bef_dup bef_swap
%type <com>		bef_mult bef_add bef_div bef_sub bef_fmult bef_fadd
%type <com>		bef_fdiv bef_fsub bef_ftoi bef_itof bef_inot bef_and
%type <com>		bef_or bef_not bef_shl bef_shr bef_print bef_fprint
%type <com>		bef_cprint bef_read bef_fread bef_sread bef_cread
%type <com>		bef_jump bef_call bef_bzero bef_bnzro bef_bpos 
%type <com>		bef_bneg bef_ret bef_sprint bef_exch bef_loadw
%type <com>		bef_loadb bef_storew bef_storeb
%type <com>		partypa partypb
%type <integer>		int_exp offset
%type <numeric>		dbl_exp dbl_exp_x

%start file

%%

file:		{loadaddr=0;line=1;if(loaddebug&&asmlog)fprintf(asmlog,"\n");} items
	;

items:		items item
	|	item
	;

item:		befehl {
	  switch (($1.opcode) & 0x03) {
	  case 0:
	    if (loaddebug&&asmlog) {
	      fprintf (asmlog, " %04x:   %02x %15s %s\n",
		       loadaddr, (int) $1.opcode, "",
		       opnames[($1.opcode>>2)]);
	    }
	    PUTB (loadaddr, $1.opcode);
	    loadaddr++;
	    break;

	  case 1:
	    PUTB (loadaddr,   $1.opcode);
	    PUTW (loadaddr+1, $1.parameter);

	    if (loaddebug&&asmlog) {
	      fprintf (asmlog, " %04x:   %02x %02x %02x %9s %s $%04x\n",
		       loadaddr, 
		       (int) speicher[loadaddr],
		       (int) speicher[loadaddr+1],
		       (int) speicher[loadaddr+2],
		       "",
		       opnames[($1.opcode>>2)],
		       (int) $1.parameter);
	    }
	    loadaddr += 3;
	    break;

	  case 2:
	    PUTB (loadaddr,   $1.opcode);
	    PUTD (loadaddr+1, $1.parameter);

	    if (loaddebug&&asmlog) {
	      fprintf (asmlog, 
		       " %04x:   %02x %02x %02x %02x %02x %3s %s #$%08x\n",
		       loadaddr, 
		       (int) speicher[loadaddr],
		       (int) speicher[loadaddr+1],
		       (int) speicher[loadaddr+2],
		       (int) speicher[loadaddr+3],
		       (int) speicher[loadaddr+4],
		       "",
		       opnames[($1.opcode>>2)],
		       (int) $1.parameter);
	    }
	    loadaddr += 5;
	    break;
	  }
	}
	|	DATA datalist {
	}
	|	LABDEF {
	  AddLabel ($1, loadaddr);
	  if (loaddebug&&asmlog) {
	    fprintf (asmlog, " %04x: %s\n", loadaddr, $1);
	  }
	}
	|	':' INTEGER {
	  loadaddr = $2;
	}
	|	':' ALIGN {
	  loadaddr = (loadaddr+3)&0xfffc;
	}
	;

datalist:	datalist ',' datum
	|	datum
	;

datum:		int_exp {
	  PUTD (loadaddr, $1);
	  if (loaddebug&&asmlog) {
	    fprintf (asmlog, 
		     " %04x:   %02x %02x %02x %02x        <idata> %08x\n",
		     loadaddr,
		     (int) speicher[loadaddr+0],
		     (int) speicher[loadaddr+1],
		     (int) speicher[loadaddr+2],
		     (int) speicher[loadaddr+3],
		     $1);
	  }
	  loadaddr += 4;
	}
	|	dbl_exp {
	  memcpy (&(speicher[loadaddr]), &($1), sizeof (double));
	  if (loaddebug&&asmlog) {
	    fprintf (asmlog,
		     " %04x:   %02x %02x %02x %02x        <fdata> %g\n",
		     loadaddr,
		     (int) speicher[loadaddr+0],
		     (int) speicher[loadaddr+1],
		     (int) speicher[loadaddr+2],
		     (int) speicher[loadaddr+3],
		     $1);
	    fprintf (asmlog, " %04x:   %02x %02x %02x %02x\n",
		     loadaddr + 4,
		     (int) speicher[loadaddr+4],
		     (int) speicher[loadaddr+5],
		     (int) speicher[loadaddr+6],
		     (int) speicher[loadaddr+7]);
	  }
	  loadaddr += sizeof (double);
	}
	|	STRING {
	  memcpy (&(speicher[loadaddr]), &($1), strlen ($1) + 1);

	  if (loaddebug&&asmlog) {
	    int index, imax, iter;
	    for (index=0, imax=strlen($1)+1; index<imax; index+=5) {
	      fprintf (asmlog, " %04x:   %02x ",
		       loadaddr,
		       (int)speicher[loadaddr+index]);
	      for (iter=1; iter<5; iter++) {
		if (index+iter<imax)
		  fprintf (asmlog, "%02x ", 
			   (int)speicher[loadaddr+index+iter]);
		else
		  fprintf (asmlog, "   ");
	      }
	      fprintf (asmlog, "    <sdata> ");
	      for (iter=0; iter<5; iter++) {
		if (index+iter<imax)
		  fprintf (asmlog, "%c",
			   (speicher[loadaddr+index+iter]<0x20||
			    speicher[loadaddr+index+iter]>0x7f) ?
			   '.' : speicher[loadaddr+index+iter]);
	      }
	      fprintf (asmlog, "\n");
	    }
	  }
	  loadaddr += strlen ($1) + 1;
	}
	|	'x' int_exp {
	  if (loaddebug&&asmlog) {
	    fprintf (asmlog, " %04x:   %02x %15s <bdata> %02x\n",
		     loadaddr, (int) $2, "", (int) $2);
	  }
	  PUTB (loadaddr, $2);
	  loadaddr++;
	}
	|	'w' int_exp {
	  PUTW (loadaddr, $2);

	  if (loaddebug&&asmlog) {
	    fprintf (asmlog, " %04x:   %02x %02x %12s <wdata> %04x\n",
		     loadaddr, 
		     (int) speicher[loadaddr],
		     (int) speicher[loadaddr+1],
		     "", (int) $2);
	  }
	  loadaddr += 2;
	}
	|	LABEL offset {
	  if (loaddebug&&asmlog) {
	    fprintf (asmlog, " %04x:   %02x %02x %12s <addr>  %s+%04x\n",
		     loadaddr, 0, 0, "", $1, $2);
	  }
	  RefLabel ($1, loadaddr);
	  PUTW     (loadaddr, $2);
	  loadaddr += 2;
	}
	;

befehl:		bef_load
	|	bef_loadw
	|	bef_loadb
	|	bef_store
	|	bef_storew
	|	bef_storeb
	|	bef_push
	|	bef_pop
	|	bef_dup
	|	bef_exch
	|	bef_swap
	|	bef_mult
	|	bef_add
	|	bef_div
	|	bef_sub
	|	bef_fmult
	|	bef_fadd
	|	bef_fdiv
	|	bef_fsub
	|	bef_ftoi
	|	bef_itof
	|	bef_inot
	|	bef_and
	|	bef_or
	|	bef_not
	|	bef_shl
	|	bef_shr
	|	bef_print
	|	bef_fprint
	|	bef_sprint
	|	bef_cprint
	|	bef_read
	|	bef_fread
	|	bef_sread
	|	bef_cread
	|	bef_jump
	|	bef_call
	|	bef_bzero
	|	bef_bnzro
	|	bef_bpos
	|	bef_bneg
	|	bef_ret
	;

bef_load:	LOAD partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_LOAD,0);
	}
	;

bef_loadw:	LOADW partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_LOADW,0);
	}
	;

bef_loadb:	LOADB partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_LOADB,0);
	}
	;

bef_store:	STORE partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_STORE,0);
	}
	;

bef_storew:	STOREW partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_STOREW,0);
	}
	;

bef_storeb:	STOREB partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_STOREB,0);
	}
	;

bef_push:	PUSH { $$.opcode = OPC_PUSH_STK; }
	;

bef_pop:	POP { $$.opcode = OPC_POP_STK; }
	;

bef_dup:	DUP { $$.opcode = OPC_DUP_STK; }
	;

bef_swap:	SWAP { $$.opcode = OPC_SWAP_STK; }
	;

bef_exch:	EXCH { $$.opcode = OPC_EXCH_STK; }
	;

bef_mult:	MULT partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_MULT,0);
	}
	;

bef_add:	ADD partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_ADD,0);
	}
	;

bef_div:	DIV partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_DIV,0);
	}
	;

bef_sub:	SUB partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_SUB,0);
	}
	;

bef_fmult:	FMULT { $$.opcode = OPC_FMULT_STK; }
	;

bef_fadd:	FADD { $$.opcode = OPC_FADD_STK; }
	;

bef_fdiv:	FDIV { $$.opcode = OPC_FDIV_STK; }
	;

bef_fsub:	FSUB { $$.opcode = OPC_FSUB_STK; }
	;

bef_ftoi:	FTOI partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_FTOI,0);
	}
	;

bef_itof:	ITOF partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_ITOF,0);
	}
	;

bef_inot:	INOT { $$.opcode = OPC_INOT_STK; }
	;

bef_not:	NOT { $$.opcode = OPC_NOT_STK; }
	;

bef_and:	AND partypa { 
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_AND,0);
	}
	;

bef_or:		OR partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_OR,0);
	}
	;

bef_shl:	SHL partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_SHL,0);
	}
	;

bef_shr:	SHR partypa {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_SHR,0);
	}
	;

bef_print:	PRINT { $$.opcode = OPC_PRINT_STK; }
	;

bef_fprint:	FPRINT { $$.opcode = OPC_FPRINT_STK; }
	;

bef_sprint:	SPRINT { $$.opcode = OPC_SPRINT_STK; }
	;

bef_cprint:	CPRINT { $$.opcode = OPC_CPRINT_STK; }
	;

bef_read:	READ { $$.opcode = OPC_READ_STK; }
	;

bef_fread:	FREAD { $$.opcode = OPC_FREAD_STK; }
	;

bef_sread:	SREAD { $$.opcode = OPC_SREAD_STK; }
	;

bef_cread:	CREAD { $$.opcode = OPC_CREAD_STK; }
	;

bef_jump:	JUMP partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_JUMP,0);
	}
	;

bef_call:	CALL partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_CALL,0);
	}
	;

bef_bzero:	BZERO partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_BZERO,0);
	}
	;

bef_bnzro:	BNZRO partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_BNZRO,0);
	}
	;

bef_bpos:	BPOS partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_BPOS,0);
	}
	;

bef_bneg:	BNEG partypb {
	  memcpy (&($$), &($2), sizeof (struct _t_com));
	  $$.opcode   |= MKCODE(OPC_BNEG,0);
	}
	;

bef_ret:	RET { $$.opcode = OPC_RET_STK; }
	;

partypa:	'#' int_exp {
	  $$.opcode    = MKCODE(0,ADR_IMM);
	  $$.parameter = $2;
	}
	|	'#' LABEL offset {
	  $$.opcode    = MKCODE(0,ADR_IMM);
	  $$.parameter = (unsigned short) $3;
	  RefLabel ($2, loadaddr+1);
	}
	|	int_exp {
	  $$.opcode    = MKCODE(0,ADR_MEM);
	  $$.parameter = (unsigned short) $1;
	}
	|	LABEL offset {
	  $$.opcode    = MKCODE(0,ADR_MEM);
	  $$.parameter = (unsigned short) $2;
	  RefLabel ($1, loadaddr+1);
	}
	|	/* leer */ {
	  $$.opcode    = MKCODE(0,ADR_STK);
	}
	;

partypb:	int_exp {
	  $$.opcode    = MKCODE(0,ADR_MEM);
	  $$.parameter = (unsigned short) $1;
	}
	|	LABEL offset {
	  $$.opcode    = MKCODE(0,ADR_MEM);
	  $$.parameter = (unsigned short) $2;
	  RefLabel ($1, loadaddr+1);
	}
	|	/* leer */ {
	  $$.opcode    = MKCODE(0,ADR_STK);
	}
	;

offset:		'+' int_exp {
	  $$ = $2;
	}
	|	'-' int_exp {
	  $$ = -$2;
	}
	|	/* leer */ {
	  $$ = 0;
	}
	;

int_exp:	'(' int_exp ')' {
	  $$ = $2;
	}
	|	int_exp '*' int_exp {
	  $$ = $1 * $3;
	}
	|	int_exp '+' int_exp {
	  $$ = $1 + $3;
	}
	|	int_exp '-' int_exp {
	  $$ = $1 - $3;
	}
	|	INTEGER
	;

dbl_exp:	'(' dbl_exp ')' {
	  $$ = $2;
	}
	|	dbl_exp '*' dbl_exp_x {
	  $$ = $1 * $3;
	}
	|	dbl_exp '+' dbl_exp_x {
	  $$ = $1 + $3;
	}
	|	dbl_exp '-' dbl_exp_x {
	  $$ = $1 - $3;
	}
	|	dbl_exp '/' dbl_exp_x {
	  $$ = $1 / $3;
	}
	|	NUMERIC
	;

dbl_exp_x:	'(' dbl_exp_x ')' {
	  $$ = $2;
	}
	|	dbl_exp_x '*' dbl_exp_x {
	  $$ = $1 * $3;
	}
	|	dbl_exp_x '+' dbl_exp_x {
	  $$ = $1 + $3;
	}
	|	dbl_exp_x '-' dbl_exp_x {
	  $$ = $1 - $3;
	}
	|	NUMERIC
	|	INTEGER {
	  $$ = (double) $1;
	}
	;

%%

yyerror (char *s)
{
  fprintf (stderr, "%s in %s line %d\n", s, filename, line);
}
